/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   locktab.c
 *    lock table implementation
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2021 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>

#include "mpsdef.h"
#include "locktab.h"
#include "shmmgr.h"
#include "mref.h"

void locktab_init(void)
{
    
    return;
    
    //printf ("locktab_init():  locktab at '%p'\r\n", shm_config->hdr->locktab_head);
}

short locktab_get_sem(void)
{
    time_t start_secs = time (NULL);
    time_t end_secs = time (NULL);

    /* wait 5 seconds for the semaphore */
    while ((end_secs - start_secs) < 5) {

        if (shm_config->hdr->locktab_sem == 0) {
            shm_config->hdr->locktab_sem = pid;
            return TRUE;
        }

        end_secs = time (NULL);
        sleep (1);
        
    }
    
    return FALSE;
}

void locktab_release_sem(void)
{

    if (shm_config->hdr->locktab_sem == pid) {
        shm_config->hdr->locktab_sem = 0;
    }
    
    return;
}

void lock(char *lockarg, long time_out, char type)
{
    char *key = &(lockarg[1]);
    char a = lockarg[0];
    
    if (shm_config == NULL) {
        fprintf (stderr, "lock:  global LOCK operation attemped before shared memory available.\r\n");
        return;
    }

    if (locktab_get_sem () == TRUE) {
    
        switch (a) {

            case '+':
                locktab_increment (key, time_out);
                break;
            
            case '-':
                locktab_decrement (key, time_out);
                break;
                
            case SP:
                locktab_unlock_all ();
                locktab_lock (key, time_out);
                break;

            default:
                break;

        }

        locktab_release_sem ();

    }
    else {
        /* could not acquire the locktab semaphore */
        fprintf (stderr, "FATAL:  could not acquire lock table semaphore\r\n");
        exit (1);
    }

    return;
    
}

void locktab_increment(char *key, long timeout)
{
    
    switch (timeout) {
            
        case -1: /* blocking lock (no timeout) */
        
            for (;;) {
                
                if (locktab_insert (key) != NULL) {
                    return;
                }
                else {
                    sleep (1);
                }
                
            }

            
        case 0: /* lock that returns immediately */

            if (locktab_insert (key) != NULL) {
                test = 1;
            }
            else {
                test = 0;
            }

            break;

        case 1: /* special case: lock with 1-second timeout */

            if (locktab_insert (key) != NULL) {
                test = 1;            
            }
            else {
                sleep (1);

                if (locktab_insert (key) != NULL) {
                    test = 1;
                }
                else {
                    test = 0;
                }
            }

            return;
            
        default: /* lock with timeout */
        {
            time_t start_secs;
            time_t end_secs;
            time_t elapsed;
            locktab_ent_t *lck = NULL;
            
            start_secs = time (NULL);

            for (;;) {


                lck = locktab_insert (key);
                
                end_secs = time (NULL);
                elapsed = end_secs - start_secs;

                if (lck != NULL) {
                    test = 1;
                    return;
                }
                
                if (elapsed >= timeout) {

                    if (lck == NULL) {
                        test = 0;          
                    }
                    else {
                        test = 1;
                    }

                    return;
                    
                }

                sleep (1); /* prevent CPU pegging */
                
            } /* timeout loop */

            return;
            
        } /* lock with timeout */
    } /* switch (timeout) */
            
            
}

void locktab_decrement(char *key, long timeout)
{

    locktab_ent_t *lck = locktab_find (key);

    if (lck != NULL) {
        if (lck->ct > 0) lck->ct--;

        if (lck->ct == 0) {
            lck->owner_job = 0;
            strcpy (lck->namespace, "<REUSABLE>");

            ssvn_lock_remove (lck->nref);
        }

        if (lck->owner_job != 0) {
            ssvn_lock_add (lck->nref, lck->owner_job, lck->ct);
        }
        
        if (timeout > -1) test = 1;
    }
    else {
        if (timeout > -1) test = 0;
    }
    
}

void locktab_unlock_all(void)
{
    locktab_ent_t *lck;
    
    for (lck = shm_config->hdr->locktab_head; lck != NULL; lck = lck->next) {

        if (lck->owner_job == pid) {

            lck->ct = 0;
            lck->owner_job = 0;
            strcpy (lck->namespace, "<REUSABLE>");

            ssvn_lock_remove (lck->nref);
            
        }

    }       
    
}

void locktab_lock(char *key, long timeout)
{
    locktab_increment (key, timeout);
}

locktab_ent_t *locktab_find(char *key)
{

    locktab_ent_t *lck;
    char chk_ns[255];

    if (key[1] == '%') {
        snprintf (chk_ns, 255, "SYSTEM");
    }
    else {
        snprintf (chk_ns, 255, nsname);
    }

    for (lck = shm_config->hdr->locktab_head; lck != NULL; lck = lck->next) {

        if ((stcmp (lck->nref, key) == 0) && (strcmp (lck->namespace, chk_ns) == 0)) {

            if (lck->owner_job != pid) {
                return (locktab_ent_t *) NULL;
            }
            else {
                return lck;
            }
            
        }
        
    }

    return (locktab_ent_t *) NULL;
    
}

locktab_ent_t *locktab_insert(char *key)
{
    locktab_ent_t *l;
    char chk_ns[255];

    if (key[1] == '%') {
        snprintf (chk_ns, 255, "SYSTEM");
    }
    else {
        snprintf (chk_ns, 255, "%s", nsname);
    }
    
    for (l = shm_config->hdr->locktab_head; l != NULL; l = l->next) {

        if ((stcmp (l->nref, key) == 0) && (strcmp (l->namespace, chk_ns) == 0)) {

            /* nref already owned by another job */
            if ((l->owner_job != pid) && (l->ct > 0)) {
                return NULL;
            }
            else {
                /* increment the lock and return */
                l->ct++;

                /* if this was a lock with a counter of zero belonging to another pid,
                 * re-use it and take ownership of it.
                 */

                if (l->owner_job != pid) l->owner_job = pid;

                ssvn_lock_add (l->nref, l->owner_job, l->ct);
                
                return l;
            }
            
        }

    }

    /* no lock exists for key: this is a new insert */            
    l = (locktab_ent_t *) shm_alloc (sizeof (locktab_ent_t));
    NULLPTRCHK(l,"locktab_insert");

    stcpy (l->nref, key);
    snprintf (l->namespace, 255, "%s", chk_ns);
        
    l->owner_job = pid;
    l->ct = 1;
    
    l->next = shm_config->hdr->locktab_head;
    shm_config->hdr->locktab_head = l;

    ssvn_lock_add (l->nref, l->owner_job, l->ct);
    
    return l;
}


int locktab_count(char *key)
{
    locktab_ent_t *l;
    int ct = 0;

    for (l = shm_config->hdr->locktab_head; l != NULL; l = l->next) {
        if (stcmp (l->nref, key) == 0) ct++;
    }

    return ct;
}

unsigned long locktab_pages(void)
{

    locktab_ent_t *l;
    unsigned int ct = 0;
    unsigned long bytes = 0;
    unsigned long pages = 0;
    float extra;
    
    for (l = shm_config->hdr->locktab_head; l != NULL; l = l->next) {
        ct++;
        bytes += sizeof (locktab_ent_t);
    }

    pages = bytes / PG_SIZE;
    extra = bytes % PG_SIZE;
    
    if (extra > 0) {
        pages++;
    }

    return pages;
  
}

unsigned long locktab_bytes(void)
{

    locktab_ent_t *l;
    unsigned int ct = 0;
    unsigned long bytes = 0;

    for (l = shm_config->hdr->locktab_head; l != NULL; l = l->next) {
        ct++;
        bytes += sizeof (locktab_ent_t);
    }

    return bytes;

}

void locktab_dump(void)
{
    
    locktab_ent_t *l;
    unsigned long ct = 0;

    freem_ref_t *r = (freem_ref_t *) malloc (sizeof (freem_ref_t));
    NULLPTRCHK(r,"locktab_dump");

    char *ref_ext = (char *) malloc (STRLEN * sizeof (char));
    NULLPTRCHK(ref_ext,"locktab_dump");
    
    printf ("%-20s%-20s%-20s%s\r\n", "NAMESPACE", "PID", "COUNT", "KEY");
    printf ("%-20s%-20s%-20s%s\r\n", "---------", "---", "-----", "---");

    if (shm_config->hdr->locktab_head == NULL) {
        printf ("\r\n*** lock table empty ***\r\n");
        free (r);
        return;
    }

    
    for (l = shm_config->hdr->locktab_head; l != NULL; l = l->next) {
        mref_init (r, MREF_RT_GLOBAL, "");
        internal_to_mref (r, l->nref);
        mref_to_external (r, ref_ext);
        
        printf ("%-20s%-20d%-20d%s\r\n", l->namespace, l->owner_job, l->ct, ref_ext);
        ct++;
    }

    printf ("\r\n\tLOCK table entries:                   %ld\r\n", ct);
    printf (    "\tShared memory pages:                  %ld\r\n", locktab_pages ());
    printf (    "\tShared memory bytes:                  %ld\r\n", locktab_bytes ());
    if (shm_config->hdr->locktab_sem != 0) {
        printf ("\tLOCK table semaphore:                 owned by %d\r\n", shm_config->hdr->locktab_sem);
    }
    else {
        printf ("\tLOCK table semaphore:                 <UNCLAIMED>\r\n");
    }

    free (r);
    free (ref_ext);
    return;
    
}
